/*
 * PF_UNIX socket family helpers
 *
 * Grigoriy Sitkarev, <sitkarev@unixkomi.ru>
 */

#include <sys/fcntl.h>
#include <sys/socket.h>
#include <sys/stat.h>

#include <linux/un.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

#include "common.h"
#include "macros.h"
#include "xmalloc.h"
#include "sock-unix.h"

//#define DEBUG_SOCK_UNIX

#ifdef DEBUG_SOCK_UNIX
#define debug(fmt, arg...) fprintf(stderr, "debug: %s: "fmt CRLF, \
					__FUNCTION__, ##arg)
#else
#define debug(fmt, arg...) do {} while(0)
#endif

#define warn(fmt, arg...) fprintf(stderr, "warn: %s: "fmt CRLF, \
					__FUNCTION__, ##arg)

#define verbose(fmt, arg...) fprintf(stderr, "verbose: %s: "fmt CRLF, \
					__FUNCTION__, ##arg)

#define info(fmt, arg...) fprintf(stderr, "info: %s: "fmt CRLF, \
					__FUNCTION__, ##arg)

#define error(eval, fmt, arg...)			\
do {							\
	fprintf(stderr, "error: %s: "fmt CRLF,		\
				__FUNCTION__, ##arg);   \
	exit(eval);					\
} while (0)

/* create client connection */
int sock_unix_connect(const char *path)
{
	int res, sock;
	struct sockaddr_un addr;

	if (!path)
		error(1, "path NULL");

	memset(&addr, 0, sizeof(addr));
	addr.sun_family = PF_UNIX;
	xstrncpy(addr.sun_path, path, UNIX_PATH_MAX);
	sock = socket(PF_UNIX, SOCK_STREAM, 0);
	if (sock == -1) {
		verbose("sock_unix_connect: socket() '%s': %s", path, strerror(errno));
		return -1;
	}
	res = connect(sock, (struct sockaddr *)&addr, sizeof(addr));
	if (res == -1) {
		do {
			res = close(sock);
		} while (res == -1 && errno == EINTR);
		verbose("sock_unix_connect: connect() '%s': %s", path, strerror(errno));
		return -1;
	}
	return sock;
}

/* sets nonblocking mode on socket descriptor */
void sock_nonblock(int sock)
{
	int res;
	int flags = 0;

	if ((flags = fcntl(sock, F_GETFL)) == -1)
		warn("sock_nonblock: F_GETFL sock=%d res=%d: %s",
				sock, flags, strerror(errno));
	if ((res = fcntl(sock, F_SETFL, flags | O_NONBLOCK)))
		warn("sock_nonblock: F_SETFL sock=%d res=%d: %s",
				sock, res, strerror(errno));
}

/* sets blocking mode on socket descriptor */
void sock_block(int sock)
{
	int res;
	int flags = 0;

	if ((flags = fcntl(sock, F_GETFL)) == -1)
		warn("sock_block: F_GETFL sock=%d res=%d: %s",
				sock, flags, strerror(errno));
	if ((res = fcntl(sock, F_SETFL, flags &= ~O_NONBLOCK)))
		warn("sock_block: F_SETFL sock=%d res=%d: %s",
				sock, res, strerror(errno));
}

/* closes listening PF_UNIX socket */
int sock_unix_close(int sock)
{
	int res;
	socklen_t slen;
	struct sockaddr_un addr;
	struct stat st;

	if (sock < 0)
		error(1, "sock_unix_unlink: sock=%d", sock);

	slen = sizeof(struct sockaddr_un);
	res = getsockname(sock, (struct sockaddr *)&addr, &slen);
	if (res) {
		verbose("sock_unix_unlink: getsockname()=%d: %s",
				res, strerror(errno));
		return -1;
	}
	if (addr.sun_family != PF_UNIX || !addr.sun_path[0])
		return -1;
	do {
		res = close(sock);
	} while (res == -1 && errno == EINTR);
	res = stat(addr.sun_path, &st);
	if (res) {
		verbose("sock_unix_unlink: stat()=%d: %s",
				res, strerror(errno));
		return -1;
	}
	if (((S_IFSOCK & st.st_mode) == 0) || unlink(addr.sun_path)) {
		verbose("sock_unix_unlink: unlink() failed or not socket");
		return -1;
	}
	return 0;
}

/* creates PF_UNIX socket ready to accept connections */
int sock_unix_listen(const char *path)
{
	int res, sock;
	struct sockaddr_un addr;

	if (!path)
		error(1, "sock_unix_listen: path NULL");

	memset(&addr, 0, sizeof(addr));
	addr.sun_family = PF_UNIX;
	xstrncpy(addr.sun_path, path, UNIX_PATH_MAX);
	sock = socket(PF_UNIX, SOCK_STREAM, 0);
	if (sock == -1) {
		verbose("sock_unix_listen: PF_UNIX '%s' socket error: %s",
				path, strerror(errno));
		return -1;
	}
	res = bind(sock, (struct sockaddr *) &addr, sizeof(addr));
	if (res == -1) {
		verbose("sock_unix_listen: can't bind() to '%s': %s",
				path, strerror(errno));
		do {
			res = close(sock);
		} while (res == -1 && errno == EINTR);
		return -1;
	}
	res = listen(sock, 10);
	if (res == -1) {
		verbose("sock_unix_listen: can't listen() on '%s': %s",
				path, strerror(errno));
		do {
			res = close(sock);
		} while (res == -1 && errno == EINTR);
		res = unlink(addr.sun_path);
		if (res)
			verbose("sock_unix_listen: can't unlink '%s': %s",
					addr.sun_path, strerror(errno));
		return -1;
	}
	return sock;
}
